# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'canopy_cli': 'src/canopy_cli',
 'canopy_cli.data_loader': 'src/canopy_cli/data_loader',
 'canopy_server': 'src/canopy_server',
 'canopy_server.models.v1': 'src/canopy_server/models/v1'}

packages = \
['canopy',
 'canopy.chat_engine',
 'canopy.chat_engine.history_pruner',
 'canopy.chat_engine.prompt_builder',
 'canopy.chat_engine.query_generator',
 'canopy.context_engine',
 'canopy.context_engine.context_builder',
 'canopy.knowledge_base',
 'canopy.knowledge_base.chunker',
 'canopy.knowledge_base.record_encoder',
 'canopy.knowledge_base.reranker',
 'canopy.llm',
 'canopy.models',
 'canopy.tokenizer',
 'canopy.utils',
 'canopy_cli',
 'canopy_cli.data_loader',
 'canopy_server',
 'canopy_server.models.v1']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.92.0,<0.93.0',
 'gunicorn>=21.2.0,<22.0.0',
 'jsonschema>=4.2.0,<5.0.0',
 'openai>=1.2.3,<2.0.0',
 'pandas-stubs>=2.0.3.230814,<3.0.0.0',
 'pinecone-client>=2.2.2,<3.0.0',
 'pinecone-datasets>=0.6.2,<0.7.0',
 'pinecone-text>=0.7.1,<0.8.0',
 'prompt-toolkit>=3.0.39,<4.0.0',
 'pydantic>=1.10.7,<2.0.0',
 'python-dotenv>=1.0.0,<2.0.0',
 'sentencepiece>=0.1.99,<0.2.0',
 'sse-starlette>=1.6.5,<2.0.0',
 'tenacity>=8.2.1,<9.0.0',
 'tiktoken>=0.3.3,<0.4.0',
 'tokenizers>=0.15.0,<0.16.0',
 'tqdm>=4.66.1,<5.0.0',
 'transformers>=4.35.2,<5.0.0',
 'types-jsonschema>=4.2.0,<5.0.0',
 'types-pyyaml>=6.0.12.12,<7.0.0.0',
 'types-tqdm>=4.61.0,<5.0.0',
 'uvicorn>=0.20.0,<0.21.0']

extras_require = \
{'cohere': ['cohere>=4.37']}

entry_points = \
{'console_scripts': ['canopy = canopy_cli.cli:cli']}

setup_kwargs = {
    'name': 'canopy-sdk',
    'version': '0.5.0',
    'description': 'Retrieval Augmented Generation (RAG) framework and context engine powered by Pinecone',
    'long_description': '# Canopy\n\n<p align="center">\n<a href="https://pypi.org/project/canopy-sdk" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/canopy-sdk" alt="Supported Python versions">\n</a>\n<a href="https://pypi.org/project/canopy-sdk" target="_blank">\n    <img src="https://img.shields.io/pypi/v/canopy-sdk?label=pypi%20package" alt="Package version">\n</a>\n</p>\n\n\n**Canopy** is an open-source Retrieval Augmented Generation (RAG) framework and context engine built on top of the Pinecone vector database. Canopy enables you to quickly and easily experiment with and build applications using RAG. Start chatting with your documents or text data with a few simple commands.   \n\nCanopy takes on the heavy lifting for building RAG applications: from chunking and embedding your text data to chat history management, query optimization, context retrieval (including prompt engineering), and augmented generation. \n\nCanopy provides a configurable built-in server so you can effortlessly deploy a RAG-powered chat application to your existing chat UI or interface. Or you can build your own, custom RAG application using the Canopy library. \n\nCanopy lets you evaluate your RAG workflow with a CLI based chat tool. With a simple command in the Canopy CLI you can interactively chat with your text data and compare RAG vs. non-RAG workflows side-by-side. \n\nCheck out our [blog post](https://pinecone.io/blog/canopy-rag-framework) to learn more, or see a quick [tutorial here](https://www.youtube.com/watch?v=dVGPglKh80Y). \n\n## RAG with Canopy\n\n![](.readme-content/rag_flow.png)\n\nCanopy implements the full RAG workflow to prevent hallucinations and augment your LLM with your own text data.\n\nCanopy has two flows: knowledge base creation and chat. In the knowledge base creation flow, users upload their documents and transform them into meaningful representations stored in Pinecone\'s Vector Database. In the chat flow, incoming queries and chat history are optimized to retrieve the most relevant documents, the knowledge base is queried, and a meaningful context is generated for the LLM to answer.\n\n## What\'s inside the box?\n\n1. **Canopy Core Library** - The library has 3 main classes that are responsible for different parts of the RAG workflow:\n    * **ChatEngine** - Exposes a chat interface to interact with your data. Given the history of chat messages, the `ChatEngine` formulates relevant queries to the `ContextEngine`, then uses the LLM to generate a knowledgeable response.\n    * **ContextEngine**  - Performs the “retrieval” part of RAG. The `ContextEngine` utilizes the underlying `KnowledgeBase` to retrieve the most relevant documents, then formulates a coherent textual context to be used as a prompt for the LLM. \n    * **KnowledgeBase** - Manages your data for the RAG workflow. It automatically chunks and transforms your text data into text embeddings, storing them in a Pinecone vector database. Given a text query - the `KnowledgeBase` will retrieve the most relevant document chunks from the database. \n\n\n> More information about the Core Library usage can be found in the [Library Documentation](docs/library.md)\n\n2. **Canopy Server** - This is a webservice that wraps the **Canopy Core** library and exposes it as a REST API. The server is built on top of FastAPI, Uvicorn and Gunicorn and can be easily deployed in production. \nThe server also comes with a built-in Swagger UI for easy testing and documentation. After you [start the server](#3-start-the-canopy-server), you can access the Swagger UI at `http://host:port/docs` (default: `http://localhost:8000/docs`)\n\n 3. **Canopy CLI** - A built-in development tool that allows users to swiftly set up their own Canopy server and test its configuration.  \nWith just three CLI commands, you can create a new Canopy server, upload your documents to it, and then interact with the Chatbot using a built-in chat application directly from the terminal. The built-in chatbot also enables comparison of RAG-infused responses against a native LLM chatbot.\n\n## Setup\n\n0. set up a virtual environment (optional)\n```bash\npython3 -m venv canopy-env\nsource canopy-env/bin/activate\n```\n> More information about virtual environments can be found [here](https://docs.python.org/3/tutorial/venv.html)\n\n1. install the package\n```bash\npip install canopy-sdk\n```\n\n2. Set up the environment variables\n\n```bash\nexport PINECONE_API_KEY="<PINECONE_API_KEY>"\nexport PINECONE_ENVIRONMENT="<PINECONE_ENVIRONMENT>"\nexport OPENAI_API_KEY="<OPENAI_API_KEY>"\nexport INDEX_NAME="<INDEX_NAME>"\n```\n\n<details>\n<summary><b><u>CLICK HERE</u></b> for more information about the environment variables \n\n<br /> \n</summary>\n\n| Name                  | Description                                                                                                                 | How to get it?                                                                                                                                                               |\n|-----------------------|-----------------------------------------------------------------------------------------------------------------------------|------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|\n| `PINECONE_API_KEY`    | The API key for Pinecone. Used to authenticate to Pinecone services to create indexes and to insert, delete and search data | Register or log into your Pinecone account in the [console](https://app.pinecone.io/). You can access your API key from the "API Keys" section in the sidebar of your dashboard |\n| `PINECONE_ENVIRONMENT`| Determines the Pinecone service cloud environment of your index e.g `west1-gcp`, `us-east-1-aws`, etc                       | You can find the Pinecone environment next to the API key in [console](https://app.pinecone.io/)                                                                             |\n| `OPENAI_API_KEY`      | API key for OpenAI. Used to authenticate to OpenAI\'s services for embedding and chat API                                    | You can find your OpenAI API key [here](https://platform.openai.com/account/api-keys). You might need to login or register to OpenAI services                                |\n| `ANYSCALE_API_KEY`    | API key for Anyscale. Used to authenticate to Anyscale Endpoints for open source LLMs                                    | You can register Anyscale Endpoints and find your API key [here](https://app.endpoints.anyscale.com/)\n| `CO_API_KEY`   | API key for Cohere. Used to authenticate to Cohere services for embedding                                           | You can find more information on registering to Cohere [here](https://cohere.com/pricing)\n| `INDEX_NAME`          | Name of the Pinecone index Canopy will underlying work with                                                                  | You can choose any name as long as it follows Pinecone\'s [restrictions](https://support.pinecone.io/hc/en-us/articles/11729246212637-Are-there-restrictions-on-index-names-#:~:text=There%20are%20two%20main%20restrictions,and%20emojis%20are%20not%20supported.)                                                                                       |\n| `CANOPY_CONFIG_FILE` | The path of a configuration yaml file to be used by the Canopy server. | Optional - if not provided, default configuration would be used |\n</details>\n\n\n3. Check that installation is successful and environment is set, run:\n```bash\ncanopy\n```\n\nOutput should be similar to this:\n\n```bash\nCanopy: Ready\n\nUsage: canopy [OPTIONS] COMMAND [ARGS]...\n# rest of the help message\n```\n\n## Quickstart\n\nIn this quickstart, we will show you how to use the **Canopy** to build a simple question answering system using RAG (retrieval augmented generation).\n\n### 1. Create a new **Canopy** Index\n\nAs a one-time setup, Canopy needs to create a new Pinecone index that is configured to work with Canopy, just run:\n\n```bash\ncanopy new\n```\n\nAnd follow the CLI instructions. The index that will be created will have a prefix `canopy--<INDEX_NAME>`.   \nYou only have to do this process once for every Canopy index you want to create.\n\n> To learn more about Pinecone indexes and how to manage them, please refer to the following guide: [Understanding indexes](https://docs.pinecone.io/docs/indexes)\n\n### 2. Uploading data\n\nYou can load data into your Canopy index using the command:\n\n```bash\ncanopy upsert /path/to/data_directory\n# or\ncanopy upsert /path/to/data_directory/file.parquet\n# or\ncanopy upsert /path/to/data_directory/file.jsonl\n# or\ncanopy upsert /path/to/directory_of_txt_files/\n# ...\n```\n\nCanopy supports files in `jsonl`, `parquet` and `csv` formats. Additionally, you can load plaintext data files in `.txt` format. In this case, each file will be treated as a single document. The document id will be the filename, and the source will be the full path of the file. \n\n> **Note**: Document fields are used in the RAG flow and should comply with the following schema:\n\n```bash\n+----------+--------------+--------------+---------------+\n| id(str)  | text(str)    | source       | metadata      |\n|          |              | Optional[str]| Optional[dict]|\n|----------+--------------+--------------+---------------|\n| "id1"    | "some text"  | "some source"| {"key": "val"}|\n+----------+--------------+--------------+---------------+\n\n# id       - unique identifier for the document\n#\n# text     - the text of the document, in utf-8 encoding.\n#\n# source   - the source of the document, can be any string, or null.\n#            ** this will be used as a reference in the generated context. **\n#\n# metadata - optional metadata for the document, for filtering or additional context.\n#            Dict[str, Union[str, int, float, List[str]]]\n```\n\n[This notebook](https://colab.research.google.com/github/pinecone-io/examples/blob/master/learn/generation/canopy/00-canopy-data-prep.ipynb) shows how you create a dataset in this format, Follow the instructions in the CLI when you upload your data.\n\n### 3. Start the Canopy server\n\nThe Canopy server exposes Canopy\'s functionality via a REST API. Namely, it allows you to upload documents, retrieve relevant docs for a given query, and chat with your data. The server exposes a `/chat.completion` endpoint that can be easily integrated with any chat application.\nTo start the server, run:\n\n```bash\ncanopy start\n```\nNow, you should be prompted with the following standard Uvicorn message:\n```\n...\n\nINFO:     Uvicorn running on http://0.0.0.0:8000 (Press CTRL+C to quit)\n```\n**That\'s it!** you can now start using the **Canopy** server with any chat application that supports a `/chat.completion` endpoint.\n\n> The canopy start command will keep the terminal occupied (recommended use). \n> If you want to run the server in the background, you can use the following command - **```nohup canopy start &```**\n\n\n### Stopping the server \nTo stop the server, simply press `CTRL+C` in the terminal where you started it.\n\nIf you have started the server in the background, you can stop it by running:\n\n```bash\ncanopy stop\n```\n\n## Evaluation chat tool\n\nCanopy\'s CLI comes with a built-in chat app that allows you to interactively chat with your text data and compare RAG vs. non-RAG workflows side-by-side to evaluate the results\n\nIn a new terminal window, set the [required environment variables](#setup) then run:\n\n```bash \ncanopy chat\n```\n\nThis will open a chat interface in your terminal. You can ask questions and the RAG-infused chatbot will try to answer them using the data you uploaded.\n\nTo compare the chat response with and without RAG use the `--no-rag` flag\n\n> **Note**: This method is only supported with OpenAI at the moment.\n\n```bash\ncanopy chat --no-rag\n```\n\nThis will open a similar chat interface window, but will show both the RAG and non-RAG responses side-by-side.\n\n## Considerations\n\n* Canopy currently only supports OpenAI as the backend for both the embedding model and the LLM. Rate limits and pricing set by OpenAI will apply.  \n* More integrations will be supported in the near future.\n\n## Advanced usage\n\n### Migrating an existing OpenAI application to **Canopy**\n\nIf you already have an application that uses the OpenAI API, you can migrate it to **Canopy** by simply changing the API endpoint to `http://host:port/v1`, for example with the default configuration:\n\n```python\nimport openai\n\nopenai.api_base = "http://localhost:8000/v1"\n\n# now you can use the OpenAI API as usual\n```\n\nor without global state change:\n\n```python\nimport openai\n\nopenai_response = openai.Completion.create(..., api_base="http://localhost:8000/v1")\n```\n\n### Running Canopy server in production\n\nCanopy is using FastAPI as the web framework and Uvicorn as the ASGI server. It is recommended to use Gunicorn as the production server, mainly because it supports multiple worker processes and can handle multiple requests in parallel, more details can be found [here](https://www.uvicorn.org/deployment/#using-a-process-manager).\n\nTo run the canopy server for production, please run:\n\n```bash\ngunicorn canopy_cli.app:app --worker-class uvicorn.workers.UvicornWorker --bind 0.0.0.0:8000 --workers <number of desired worker processes>\n```\n\nThe server interacts with services like Pinecone and OpenAI using your own authentication credentials. When deploying the server on a public web hosting provider, it is recommended to enable an authentication mechanism, so that your server would only take requests from authenticated users.\n',
    'author': 'Relevance Team',
    'author_email': 'relevance@pinecone.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, List, Union, Dict, Literal

from pydantic import BaseModel, Field, validator, Extra

Metadata = Dict[str, Union[str, int, float, List[str]]]


# ----------------- Context Engine models -----------------


class Query(BaseModel):
    text: str = Field(description="The query text.")
    namespace: str = Field(
        default="",
        description="The namespace of the query. To learn more about namespaces, see https://docs.pinecone.io/docs/namespaces",  # noqa: E501
    )
    metadata_filter: Optional[dict] = Field(
        default=None,
        description="A Pinecone metadata filter, to learn more about metadata filters, see https://docs.pinecone.io/docs/metadata-filtering",  # noqa: E501
    )
    top_k: Optional[int] = Field(
        default=None,
        description="The number of results to return."
    )
    query_params: dict = Field(
        default_factory=dict,
        description="Pinecone Client additional query parameters."
    )


class Document(BaseModel):
    id: str = Field(description="The document id.")
    text: str = Field(description="The document text.")
    source: str = Field(
        default="",
        description="The source of the document: a URL, a file path, etc."
    )
    metadata: Metadata = Field(
        default_factory=dict,
        description="The document metadata. To learn more about metadata, see https://docs.pinecone.io/docs/manage-data",  # noqa: E501
    )

    class Config:
        extra = Extra.forbid

    @validator("metadata")
    def metadata_reseved_fields(cls, v):
        if "text" in v:
            raise ValueError('Metadata cannot contain reserved field "text"')
        if "document_id" in v:
            raise ValueError('Metadata cannot contain reserved field "document_id"')
        if "source" in v:
            raise ValueError('Metadata cannot contain reserved field "source"')
        return v


class ContextContent(BaseModel, ABC):
    # Any context should be able to be represented as well formatted text.
    # In the most minimal case, that could simply be a call to `.json()`.
    @abstractmethod
    def to_text(self, **kwargs) -> str:
        pass

    def __str__(self):
        return self.to_text()


class Context(BaseModel):
    content: ContextContent
    num_tokens: int
    debug_info: dict = Field(default_factory=dict, exclude=True)

    def to_text(self, **kwargs) -> str:
        return self.content.to_text(**kwargs)


# --------------------- LLM models ------------------------


class Role(Enum):
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"


class MessageBase(BaseModel):
    role: Role = Field(description="The role of the message's author. "
                                   "Can be one of ['User', 'Assistant', 'System']")
    content: str = Field(description="The contents of the message.")

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        d["role"] = d["role"].value
        return d


Messages = List[MessageBase]


class UserMessage(MessageBase):
    role: Literal[Role.USER] = Role.USER
    content: str


class SystemMessage(MessageBase):
    role: Literal[Role.SYSTEM] = Role.SYSTEM
    content: str


class AssistantMessage(MessageBase):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT
    content: str



from typing import Optional, List, Union

from pydantic import BaseModel


class FunctionPrimitiveProperty(BaseModel):
    name: str
    type: str
    description: Optional[str] = None
    enum: Optional[List[str]] = None


class FunctionArrayProperty(BaseModel):
    name: str
    items_type: str
    # we require description for array properties
    # because the model is more struggling with them
    description: str

    def dict(self, *args, **kwargs):
        super_dict = super().dict(*args, **kwargs)
        if "items_type" in super_dict:
            super_dict["items"] = {"type": super_dict.pop("items_type")}
        return super_dict


FunctionProperty = Union[FunctionPrimitiveProperty, FunctionArrayProperty]


class FunctionParameters(BaseModel):
    required_properties: List[FunctionProperty]
    optional_properties: List[FunctionProperty] = []

    def dict(self, *args, **kwargs):
        return {
            "type": "object",
            "properties": {
                pro.name: pro.dict(exclude_none=True, exclude={"name"})
                for pro in self.required_properties + self.optional_properties
            },
            "required": [pro.name for pro in self.required_properties],
        }


class Function(BaseModel):
    name: str
    description: str
    parameters: FunctionParameters



from copy import deepcopy
from typing import List, Optional

from pinecone_text.sparse import SparseVector
from pydantic import BaseModel, Field

from canopy.models.data_models import Document, Query

# TODO: (1) consider moving this to pinecone-text
# TODO: (2) consider renaming to "Vector" or "DenseVector"
# TODO: (3) consider supporting `np.ndarray`
VectorValues = List[float]


class KBDocChunk(Document):
    document_id: str


class KBDocChunkWithScore(KBDocChunk):
    score: float


class KBEncodedDocChunk(KBDocChunk):
    values: VectorValues
    sparse_values: Optional[SparseVector] = None

    def to_db_record(self):
        metadata = deepcopy(self.metadata)
        metadata["text"] = self.text
        metadata["document_id"] = self.document_id
        metadata["source"] = self.source

        return {
            "id": self.id,
            "values": self.values,
            "metadata": metadata,
            "sparse_values": self.sparse_values,
        }


class KBQuery(Query):
    values: Optional[VectorValues] = None
    sparse_values: Optional[SparseVector] = None


class KBQueryResult(BaseModel):
    query: str
    documents: List[KBDocChunkWithScore]


class DocumentWithScore(Document):
    score: float


class QueryResult(BaseModel):
    query: str
    documents: List[DocumentWithScore]
    debug_info: dict = Field(default_factory=dict, exclude=True)



from typing import List

from .base import Chunker
from ..models import KBDocChunk
from canopy.tokenizer import Tokenizer
from ...models.data_models import Document


class TokenChunker(Chunker):
    """
    Simple chunker that splits a document into chunks (group of tokens) of a given size, using a tokenizer.
    A TokenChunker is a derived class of Chunker, which means that it can be referenced by a name
    and configured in a config file.
    """  # noqa: E501

    def __init__(self,
                 max_chunk_size: int = 256,
                 overlap: int = 30, ):
        """
        Using the global tokenizer, will set the class parameters for the TokenChunker.
        will check overlap and max_chunk_size.

        Args:
            max_chunk_size: size of the chunks, in tokens
            overlap: overlap between chunks, in tokens
        """  # noqa: E501

        # TODO: should add check for overlap not bigger than max_chunk_size
        if overlap < 0:
            cls_name = self.__class__.__name__
            raise ValueError(
                f"overlap for {cls_name} can't be negative, got: {overlap}"
            )

        if max_chunk_size <= 0:
            cls_name = self.__class__.__name__
            raise ValueError(
                f"max_chunk_size for {cls_name} must be positive, got: {max_chunk_size}"
            )

        self._tokenizer = Tokenizer()
        self._chunk_size = max_chunk_size
        self._overlap = overlap

    def chunk_single_document(self, document: Document) -> List[KBDocChunk]:
        """
        This methods takes a document and returns a list of KBDocChunks, where text is splitted
        evenly using the tokenizer. Firts the text is tokenized, then the tokens are splitted into chunks
        of a given size, with overlap between chunks.
        Last chunk is handled such that if the last chunk is smaller than the overlap, it will be removed.

        Args:
            document: document to be chunked

        Returns:
            text_chunks: list of chunks KBDocChunks from the document
        """  # noqa: E501
        tokens = self._tokenizer.tokenize(document.text)
        token_chunks = [tokens[i:i + self._chunk_size]
                        for i in range(0, len(tokens),
                                       self._chunk_size - self._overlap)]

        if len(token_chunks) == 0:
            return []

        # remove last chunk if it is smaller than overlap
        if len(token_chunks[-1]) <= self._overlap and len(token_chunks) > 1:
            token_chunks = token_chunks[:-1]

        text_chunks = [self._tokenizer.detokenize(chunk)
                       for chunk in token_chunks]
        return [KBDocChunk(id=f"{document.id}_{i}",
                           document_id=document.id,
                           text=text_chunk,
                           source=document.source,
                           metadata=document.metadata)
                for i, text_chunk in enumerate(text_chunks)]

    async def achunk_single_document(self, document: Document) -> List[KBDocChunk]:
        raise NotImplementedError()



from typing import Optional, List

from pydantic import BaseModel, Field

from canopy.models.data_models import Messages, Query, Document

# TODO: consider separating these into modules: Chat, Context, Application, etc.


class ChatRequest(BaseModel):
    model: str = Field(
        default="",
        description="The ID of the model to use. This field is ignored; instead, configure this field in the Canopy config.",  # noqa: E501
    )
    messages: Messages = Field(
        description="A list of messages comprising the conversation so far."
    )
    stream: bool = Field(
        default=False,
        description="""Whether or not to stream the chatbot's response. If set, the response is server-sent events containing [chat.completion.chunk](https://platform.openai.com/docs/api-reference/chat/streaming) objects""",  # noqa: E501
    )
    user: Optional[str] = Field(
        default=None,
        description="A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. Unused, reserved for future extensions",  # noqa: E501
    )

    class Config:
        extra = "ignore"


class ContextQueryRequest(BaseModel):
    queries: List[Query]
    max_tokens: int


class ContextResponse(BaseModel):
    content: str
    num_tokens: int


class ContextUpsertRequest(BaseModel):
    documents: List[Document]
    batch_size: int = Field(
        default=200, description="The batch size to use when uploading documents chunks to the Pinecone Index."  # noqa: E501
    )


class ContextDeleteRequest(BaseModel):
    document_ids: List[str] = Field(description="List of document IDs to delete.")


class HealthStatus(BaseModel):
    pinecone_status: str
    llm_status: str


class ChatDebugInfo(BaseModel):
    id: str
    duration_in_sec: float
    internal_model: str
    prompt_tokens: Optional[int] = None
    generated_tokens: Optional[int] = None

    def to_text(
        self,
    ):
        return self.json()


class ShutdownResponse(BaseModel):
    message: str = Field(
        default="Shutting down",
        description="Message indicating the server is shutting down.",
    )


class SuccessUpsertResponse(BaseModel):
    message: str = Field(
        default="Success",
        description="Message indicating the upsert was successful.",
    )


class SuccessDeleteResponse(BaseModel):
    message: str = Field(
        default="Success",
        description="Message indicating the delete was successful.",
    )



from itertools import zip_longest
from typing import List, Tuple

from pydantic import BaseModel

from canopy.context_engine.context_builder.base import ContextBuilder
from canopy.knowledge_base.models import QueryResult, DocumentWithScore
from canopy.tokenizer import Tokenizer
from canopy.models.data_models import Context, ContextContent


# ------------- DATA MODELS -------------

class ContextSnippet(BaseModel):
    source: str
    text: str


class ContextQueryResult(BaseModel):
    query: str
    snippets: List[ContextSnippet]


class StuffingContextContent(ContextContent):
    __root__: List[ContextQueryResult]

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']

    # In the case of StuffingContextBuilder, we simply want the text representation to
    # be a json. Other ContextContent subclasses may render into text differently
    def to_text(self, **kwargs):
        return self.json(**kwargs)


# ------------- CONTEXT BUILDER -------------

class StuffingContextBuilder(ContextBuilder):

    def __init__(self):
        self._tokenizer = Tokenizer()

    def build(self,
              query_results: List[QueryResult],
              max_context_tokens: int) -> Context:

        sorted_docs_with_origin = self._round_robin_sort(query_results)

        # Stuff as many documents as possible into the context
        context_query_results = [
            ContextQueryResult(query=qr.query, snippets=[])
            for qr in query_results]
        debug_info = {"num_docs": len(sorted_docs_with_origin)}
        content = StuffingContextContent(__root__=context_query_results)

        if self._tokenizer.token_count(content.to_text()) > max_context_tokens:
            return Context(content=StuffingContextContent(__root__=[]),
                           num_tokens=1, debug_info=debug_info)

        seen_doc_ids = set()
        for doc, origin_query_idx in sorted_docs_with_origin:
            if doc.id not in seen_doc_ids and doc.text.strip() != "":
                snippet = ContextSnippet(text=doc.text, source=doc.source)

                # try inserting the snippet into the context
                context_query_results[origin_query_idx].snippets.append(
                    snippet)
                seen_doc_ids.add(doc.id)
                # if the context is too long, remove the snippet
                if self._tokenizer.token_count(content.to_text()) > max_context_tokens:
                    context_query_results[origin_query_idx].snippets.pop()

        # remove queries with no snippets
        content = StuffingContextContent(
            __root__=[qr for qr in context_query_results if len(qr.snippets) > 0]
        )

        return Context(content=content,
                       num_tokens=self._tokenizer.token_count(content.to_text()),
                       debug_info=debug_info)

    @staticmethod
    def _round_robin_sort(
            query_results: List[QueryResult]
    ) -> List[Tuple[DocumentWithScore, int]]:
        sorted_docs_with_origin = []

        for docs_tuple in zip_longest(*[qr.documents for qr in query_results],
                                      fillvalue=None):
            for idx, doc in enumerate(docs_tuple):
                if doc is not None:
                    sorted_docs_with_origin.append((doc, idx))

        return sorted_docs_with_origin

    async def abuild(self,
                     query_results: List[QueryResult],
                     max_context_tokens: int) -> Context:
        raise NotImplementedError()



import sys
import threading
import itertools


class Spinner(object):

    def __init__(self, disable=False, force=False, stream=sys.stdout, cycle=None):
        _cycle = cycle or ['-', '/', '|', '\\']
        self.spinner_cycle = itertools.cycle(_cycle)
        self.disable = disable
        self.force = force
        self.stream = stream
        self.stop_running = None
        self.spin_thread = None

    def start(self):
        if self.disable:
            return
        if self.stream.isatty() or self.force:
            self.stop_running = threading.Event()
            self.spin_thread = threading.Thread(target=self.init_spin)
            self.spin_thread.start()

    def stop(self):
        if self.spin_thread:
            self.stop_running.set()
            self.spin_thread.join()

    def init_spin(self):
        while not self.stop_running.is_set():
            content_to_stream = next(self.spinner_cycle)
            self.stream.write(content_to_stream)
            self.stream.flush()
            self.stop_running.wait(0.25)
            self.stream.write(''.join(['\b'] * len(content_to_stream)))
            self.stream.flush()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return False
        self.stop()
        return False



from abc import ABC, abstractmethod
from typing import List

from canopy.knowledge_base.models import KBQueryResult
from canopy.utils.config import ConfigurableMixin


class Reranker(ABC, ConfigurableMixin):
    """
    Abstract class for rerankers. Rerankers take a list of KBQueryResult and return a list of KBQueryResult,
    where the results are reranked according to the reranker logic.
    Reranker is an abstract class that must be subclassed to be used,
    """  # noqa: E501

    @abstractmethod
    def rerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        pass

    @abstractmethod
    async def arerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        pass


class TransparentReranker(Reranker):
    """
    Transparent reranker that does nothing, it just returns the results as is. This is the default reranker.
    The TransparentReranker is used as a placeholder for future development "forcing" every result set to be reranked.
    """  # noqa: E501
    def rerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        """
        Returns the results as is.

        Args:
            results: A list of KBQueryResult to rerank.

        Returns:
            results: A list of KBQueryResult, same as the input.
        """  # noqa: E501
        return results

    async def arerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        return results



from typing import Tuple

from canopy.chat_engine.history_pruner.base import HistoryPruner
from canopy.models.data_models import Messages


class RecentHistoryPruner(HistoryPruner):

    def build(self,
              history: Messages,
              max_tokens: int) -> Tuple[Messages, int]:
        token_count = self._tokenizer.messages_token_count(history)
        if token_count < max_tokens:
            return history, token_count

        truncated_history = history[-self._min_history_messages:]
        token_count = self._tokenizer.messages_token_count(truncated_history)
        if token_count > max_tokens:
            raise ValueError(f"The {self._min_history_messages} most recent messages in"
                             f" history require {token_count} tokens, which exceeds the"
                             f" calculated limit for history of {max_tokens} tokens.")

        for message in reversed(history[:-self._min_history_messages]):
            token_count = self._tokenizer.messages_token_count(
                truncated_history + [message]
            )

            # If the message can fit into the remaining tokens, add it
            if token_count > max_tokens:
                break

            truncated_history.insert(0, message)

        token_count = self._tokenizer.messages_token_count(truncated_history)

        return truncated_history, token_count

    async def abuild(self,
                     full_history: Messages,
                     max_tokens: int) -> Tuple[Messages, int]:
        raise NotImplementedError



import os
import logging
import signal
import sys
import uuid

import openai
from multiprocessing import current_process, parent_process

import yaml
from dotenv import load_dotenv

from canopy.llm import BaseLLM
from canopy.tokenizer import Tokenizer
from canopy.knowledge_base import KnowledgeBase
from canopy.context_engine import ContextEngine
from canopy.chat_engine import ChatEngine
from starlette.concurrency import run_in_threadpool
from sse_starlette.sse import EventSourceResponse

from fastapi import (
    FastAPI,
    HTTPException,
    Body,
    APIRouter
)
import uvicorn
from typing import cast, Union

from canopy.models.api_models import (
    StreamingChatResponse,
    ChatResponse,
)
from canopy.models.data_models import Context, UserMessage
from .models.v1.api_models import (
    ChatRequest,
    ContextQueryRequest,
    ContextUpsertRequest,
    HealthStatus,
    ContextDeleteRequest,
    ShutdownResponse,
    SuccessUpsertResponse,
    SuccessDeleteResponse,
    ContextResponse,
)

from canopy.llm.openai import OpenAILLM
from canopy_cli.errors import ConfigError
from canopy import __version__


APIChatResponse = Union[ChatResponse, EventSourceResponse]

load_dotenv()  # load env vars before import of openai
openai.api_key = os.getenv("OPENAI_API_KEY")

APP_DESCRIPTION = """
Canopy is an open-source Retrieval Augmented Generation (RAG) framework and context engine built on top of the Pinecone vector database. Canopy enables you to quickly and easily experiment with and build applications using RAG. Start chatting with your documents or text data with a few simple commands.

Canopy provides a configurable built-in server, so you can effortlessly deploy a RAG-powered chat application to your existing chat UI or interface. Or you can build your own custom RAG application using the Canopy library.

## Prerequisites

### Pinecone API key
If you don't have a Pinecone account, you can sign up for a free Starter plan at https://www.pinecone.io/.
To find your Pinecone API key and environment log into Pinecone console (https://app.pinecone.io/). You can access your API key from the "API Keys" section in the sidebar of your dashboard, and find the environment name next to it.

### OpenAI API key
You can find your free trial OpenAI API key https://platform.openai.com/account/api-keys. You might need to log in or register for OpenAI services.
"""  # noqa: E501

API_VERSION = "v1"

# Global variables - Application
app: FastAPI = FastAPI(
    title="Canopy API",
    description=APP_DESCRIPTION,
    version=__version__,
    license_info={
        "name": "Apache 2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    },
)
openai_api_router = APIRouter()
context_api_router = APIRouter(prefix="/context")
application_router = APIRouter(tags=["Application"])

# Global variables - Engines
context_engine: ContextEngine
chat_engine: ChatEngine
kb: KnowledgeBase
llm: BaseLLM

# Global variables - Logging
logger: logging.Logger


@openai_api_router.post(
    "/chat/completions",
    response_model=None,
    responses={500: {"description": "Failed to chat with Canopy"}},  # noqa: E501
)
async def chat(
    request: ChatRequest = Body(...),
) -> APIChatResponse:
    """
    Chat with Canopy, using the LLM and context engine, and return a response.

    The request schema follows OpenAI's chat completion API schema: https://platform.openai.com/docs/api-reference/chat/create.
    Note that all fields other than `messages` and `stream` are currently ignored. The Canopy server uses the model parameters defined in the `ChatEngine` config for all underlying LLM calls.

    """  # noqa: E501
    try:
        session_id = request.user or "None"  # noqa: F841
        question_id = str(uuid.uuid4())
        logger.debug(f"Received chat request: {request.messages[-1].content}")
        answer = await run_in_threadpool(
            chat_engine.chat, messages=request.messages, stream=request.stream
        )

        if request.stream:

            def stringify_content(response: StreamingChatResponse):
                for chunk in response.chunks:
                    chunk.id = question_id
                    data = chunk.json()
                    yield data

            content_stream = stringify_content(cast(StreamingChatResponse, answer))
            return EventSourceResponse(content_stream, media_type="text/event-stream")

        else:
            chat_response = cast(ChatResponse, answer)
            chat_response.id = question_id
            return chat_response

    except Exception as e:
        logger.exception(f"Chat with question_id {question_id} failed")
        raise HTTPException(status_code=500, detail=f"Internal Service Error: {str(e)}")


@context_api_router.post(
    "/query",
    response_model=ContextResponse,
    responses={
        500: {"description": "Failed to query the knowledge base or build the context"}
    },
)
async def query(
    request: ContextQueryRequest = Body(...),
) -> ContextResponse:
    """
    Query the knowledge base for relevant context.
    The returned text may be structured or unstructured, depending on the Canopy configuration.
    Query allows limiting the context length in tokens to control LLM costs.
    This method does not pass through the LLM and uses only retrieval and construction from Pinecone DB.
    """  # noqa: E501
    try:
        context: Context = await run_in_threadpool(
            context_engine.query,
            queries=request.queries,
            max_context_tokens=request.max_tokens,
        )
        return ContextResponse(content=context.content.to_text(),
                               num_tokens=context.num_tokens)

    except Exception as e:
        logger.exception(e)
        raise HTTPException(status_code=500, detail=f"Internal Service Error: {str(e)}")


@context_api_router.post(
    "/upsert",
    response_model=SuccessUpsertResponse,
    responses={500: {"description": "Failed to upsert documents"}},
)
async def upsert(
    request: ContextUpsertRequest = Body(...),
) -> SuccessUpsertResponse:
    """
    Upsert documents into the knowledge base. Upserting is a way to add new documents or update existing ones.
    Each document has a unique ID. If a document with the same ID already exists, it is updated.

    The documents are chunked and encoded, then the resulting encoded chunks are sent to the Pinecone index in batches.
    """  # noqa: E501
    try:
        logger.info(f"Upserting {len(request.documents)} documents")
        await run_in_threadpool(
            kb.upsert, documents=request.documents, batch_size=request.batch_size
        )

        return SuccessUpsertResponse()

    except Exception as e:
        logger.exception(e)
        raise HTTPException(status_code=500, detail=f"Internal Service Error: {str(e)}")


@context_api_router.post(
    "/delete",
    response_model=SuccessDeleteResponse,
    responses={500: {"description": "Failed to delete documents"}},
)
async def delete(
    request: ContextDeleteRequest = Body(...),
) -> SuccessDeleteResponse:
    """
    Delete documents from the knowledgebase. Deleting documents is done by their unique ID.
    """  # noqa: E501
    try:
        logger.info(f"Delete {len(request.document_ids)} documents")
        await run_in_threadpool(kb.delete, document_ids=request.document_ids)
        return SuccessDeleteResponse()

    except Exception as e:
        logger.exception(e)
        raise HTTPException(status_code=500, detail=f"Internal Service Error: {str(e)}")


@application_router.get(
    "/health",
    response_model=HealthStatus,
    responses={500: {"description": "Failed to connect to Pinecone or LLM"}},
)
@app.exception_handler(Exception)
async def health_check() -> HealthStatus:
    """
    Health check for the Canopy server. This endpoint checks the connection to Pinecone and the LLM.
    """  # noqa: E501
    try:
        await run_in_threadpool(kb.verify_index_connection)
    except Exception as e:
        err_msg = f"Failed connecting to Pinecone Index {kb._index_name}"
        logger.exception(err_msg)
        raise HTTPException(
            status_code=500, detail=f"{err_msg}. Error: {str(e)}"
        ) from e

    try:
        msg = UserMessage(content="This is a health check. Are you alive? Be concise")
        await run_in_threadpool(llm.chat_completion, messages=[msg], max_tokens=5)
    except Exception as e:
        err_msg = f"Failed to communicate with {llm.__class__.__name__}"
        logger.exception(err_msg)
        raise HTTPException(
            status_code=500, detail=f"{err_msg}. Error: {str(e)}"
        ) from e

    return HealthStatus(pinecone_status="OK", llm_status="OK")


@application_router.get("/shutdown")
async def shutdown() -> ShutdownResponse:
    """
    __WARNING__: Experimental method.


    This method will shutdown the server. It is used for testing purposes, and not recommended to be used
    in production.
    This method will locate the parent process and send a SIGINT signal to it.
    """  # noqa: E501
    logger.info("Shutting down")
    proc = current_process()
    p_process = parent_process()
    pid = p_process.pid if p_process is not None else proc.pid
    if not pid:
        raise HTTPException(
            status_code=500,
            detail="Failed to locate parent process. Cannot shutdown server.",
        )
    if sys.platform == 'win32':
        kill_signal = signal.CTRL_C_EVENT
    else:
        kill_signal = signal.SIGINT
    os.kill(pid, kill_signal)
    return ShutdownResponse()


@app.on_event("startup")
async def startup():
    _init_logging()
    _init_engines()
    _init_routes(app)
    await health_check()


def _init_routes(app):
    # Include the application level router (health, shutdown, ...)
    app.include_router(application_router, include_in_schema=False)
    app.include_router(application_router, prefix=f"/{API_VERSION}")
    # Include the API without version == latest
    app.include_router(context_api_router, include_in_schema=False)
    app.include_router(openai_api_router, include_in_schema=False)
    # Include the API version in the path, API_VERSION should be the latest version.
    app.include_router(context_api_router, prefix=f"/{API_VERSION}", tags=["Context"])
    app.include_router(openai_api_router, prefix=f"/{API_VERSION}", tags=["LLM"])


def _init_logging():
    global logger

    file_handler = logging.FileHandler(
        filename=os.getenv("CE_LOG_FILENAME", "canopy.log")
    )
    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    handlers = [file_handler, stdout_handler]
    logging.basicConfig(
        format="%(asctime)s - %(processName)s - %(name)-10s [%(levelname)-8s]:  "
        "%(message)s",
        level=os.getenv("CE_LOG_LEVEL", "INFO").upper(),
        handlers=handlers,
        force=True,
    )
    logger = logging.getLogger(__name__)


def _init_engines():
    global kb, context_engine, chat_engine, llm, logger

    index_name = os.getenv("INDEX_NAME")
    if not index_name:
        raise ValueError("INDEX_NAME environment variable must be set")

    config_file = os.getenv("CANOPY_CONFIG_FILE")
    if config_file:
        _load_config(config_file)

    else:
        logger.info(
            "Did not find config file. Initializing engines with default "
            "configuration"
        )
        Tokenizer.initialize()
        kb = KnowledgeBase(index_name=index_name)
        context_engine = ContextEngine(knowledge_base=kb)
        llm = OpenAILLM()
        chat_engine = ChatEngine(context_engine=context_engine, llm=llm)

    kb.connect()


def _load_config(config_file):
    global chat_engine, llm, context_engine, kb, logger
    logger.info(f"Initializing engines with config file {config_file}")
    try:
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)
    except Exception as e:
        logger.exception(f"Failed to load config file {config_file}")
        raise ConfigError(f"Failed to load config file {config_file}. Error: {str(e)}")
    tokenizer_config = config.get("tokenizer", {})
    Tokenizer.initialize_from_config(tokenizer_config)
    if "chat_engine" not in config:
        raise ConfigError(
            f"Config file {config_file} must contain a 'chat_engine' section"
        )
    chat_engine_config = config["chat_engine"]
    try:
        chat_engine = ChatEngine.from_config(chat_engine_config)
    except Exception as e:
        logger.exception(
            f"Failed to initialize chat engine from config file {config_file}"
        )
        raise ConfigError(
            f"Failed to initialize chat engine from config file {config_file}."
            f" Error: {str(e)}"
        )
    llm = chat_engine.llm
    context_engine = chat_engine.context_engine
    kb = context_engine.knowledge_base


def start(host="0.0.0.0", port=8000, reload=False, config_file=None):
    if config_file:
        os.environ["CANOPY_CONFIG_FILE"] = config_file

    uvicorn.run("canopy_server.app:app", host=host, port=port, reload=reload, workers=0)


if __name__ == "__main__":
    start()
